from collections import namedtuple


Node = namedtuple('Node', ['id', 'lon', 'lat', 'tags'])
Way = namedtuple('Way', ['id', 'points'])
Relation = namedtuple('Relation', ['id', 'tags', 'member_info'])


Point = namedtuple('Point', ['lon', 'lat'])

Agency = namedtuple('Agency', [
    'agency_id',
    'agency_url',
    'agency_name',
    'agency_timezone'])

Route = namedtuple('Route', [
    'route_id',
    'route_short_name',
    'route_long_name',
    'route_type',
    'route_url',
    'route_color',
    'agency_id'])

Stop = namedtuple('Stop', [
    'stop_id',
    'stop_name',
    'stop_lon',
    'stop_lat',
    'route_id'])

Shape = namedtuple('Shape', [
    'shape_id',
    'shape_pt_lat',
    'shape_pt_lon',
    'shape_pt_sequence'])



from .agency_builder import build_agencies
from .route_builder import build_routes
from .stop_builder import build_stops
from .shape_builder import build_shapes


__all__ = ['build_agencies', 'build_routes', 'build_stops', 'build_shapes']


#!env python
"""
Extracts partial GTFS data from OSM file.
"""
import os
import time
import tempfile
import logging
from logging import FileHandler
from contextlib import contextmanager
from pathlib import Path

import argparse

from o2g import __version__
from o2g.gtfs import gtfs_dummy
from o2g.gtfs.gtfs_writer import GTFSWriter
from o2g.osm.exporter import TransitDataExporter


class readable_dir(argparse.Action):
    def __call__(self, parser, namespace, prospective_dir, option_string=None):
        if not os.path.isdir(prospective_dir):
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Inavlid path: {0}".format(prospective_dir))
        if os.access(prospective_dir, os.R_OK):
            setattr(namespace, self.dest, prospective_dir)
        else:
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Unreadable dir: {0}".format(prospective_dir))


class readable_file(argparse.Action):
    def __call__(self, parser, namespace, prospective_file, option_string=None):
        if not os.path.isfile(prospective_file):
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Inavlid file: {0}".format(prospective_file))
        if os.access(prospective_file, os.R_OK):
            setattr(namespace, self.dest, prospective_file)
        else:
            parser.print_usage()
            print('Try "{} --help" for help.\n'.format(parser.prog))
            parser.exit("Error: Unreadable file: {0}".format(prospective_file))


def cli():
    parser = argparse.ArgumentParser(
        prog='o2g',
        epilog='Here is a smile for you :)',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Export GTFS feed from OpenStreetMap data.')
    parser.add_argument('osmfile', metavar='OSMFILE', action=readable_file,
                        help='an OSM data file supported by osmium')
    parser.add_argument('--outdir', action=readable_dir,
                        default='.',
                        help='output directory')
    parser.add_argument('--zipfile', action='store_true',
                        default=False,
                        help='save to zipfile')
    parser.add_argument('--dummy', action='store_true',
                        default=False,
                        help='fill the missing parts with dummy data')
    parser.add_argument('--loglevel',
                        default='WARNING',
                        help='DEBUG, INFO, WARNING, ERROR or CRITICAL')
    parser.add_argument('--version',
                        action='version',
                        version='%(prog)s ' + __version__,
                        help='Show the version and exit')

    args = parser.parse_args()

    if args.loglevel:
        logging.basicConfig(level=args.loglevel)
    logging.debug('Input: %s', args.osmfile)
    logging.debug('Output: %s', args.outdir)
    logging.debug('Zip?: %s', args.zipfile or False)
    logging.debug('Dummy?: %s', args.dummy)

    main(args.osmfile, args.outdir, args.zipfile, args.dummy)


def main(osmfile, outdir, zipfile, dummy):
    start = time.time()

    with capture_logs() as logfile:
        tde = TransitDataExporter(osmfile)
        tde.process()
        logging.debug('Preprocessing took %d seconds.', (time.time() - start))

        writer = GTFSWriter()
        patched_agencies = None
        if dummy:
            dummy_data = gtfs_dummy.create_dummy_data(list(tde.routes),
                                                      list(tde.stops))
            writer.add_trips(dummy_data.trips)
            writer.add_stop_times(dummy_data.stop_times)
            writer.add_calendar(dummy_data.calendar)
            writer.add_frequencies(dummy_data.frequencies)
            patched_agencies = gtfs_dummy.patch_agencies(tde.agencies)

        if patched_agencies:
            writer.add_agencies(patched_agencies)
        else:
            writer.add_agencies(tde.agencies)
        writer.add_stops(tde.stops)
        writer.add_routes(tde.routes)
        writer.add_shapes(tde.shapes)

        writer.add_file('LICENSE', Path(__file__).parents[0] / 'ODbL-1.0.txt')
        writer.add_file('logs.txt', logfile)

    if zipfile:
        writer.write_zipped(os.path.join(outdir, zipfile))
        print('GTFS feed saved in %s' % os.path.join(outdir, zipfile))
    else:
        writer.write_unzipped(outdir)
        print('GTFS feed saved in %s' % outdir)

    logging.debug('Done in %d seconds.', (time.time() - start))


@contextmanager
def capture_logs():
    logfile = tempfile.mktemp()
    handler = FileHandler(logfile, mode='w')
    logging.getLogger().addHandler(handler)
    try:
        yield logfile
    finally:
        logging.getLogger().removeHandler(handler)


if __name__ == '__main__':
    cli()



"""Common types and tools for GTFS processing."""
import enum


class GTFSRouteType(enum.Enum):
    """Route types according to the GTFS specification."""
    Tram = 0
    Subway = 1
    Rail = 2
    Bus = 3
    Ferry = 4
    CableCar = 5
    Gondola = 6
    Funicular = 7


OSM2GTFS_ROUTE_TYPE_MAP = {
    'tram': 0,
    'light_rail': 0,
    'subway': 1,
    'rail': 2,
    'railway': 2,
    'train': 2,
    'bus': 3,
    'ex-bus': 3,
    'ferry': 4,
    'cableCar': 5,
    'gondola': 6,
    'funicular': 7
}


def map_osm_route_type_to_gtfs(route_type, default=-1):
    "Get GTFS equivalent code for the given route type."
    return OSM2GTFS_ROUTE_TYPE_MAP.get(route_type, default)



"""Utilities for writing GTFS feeds."""
import os
import io
import csv
import shutil
import zipfile
from collections import OrderedDict


class GTFSWriter(object):
    """GTFS feed writer."""
    def __init__(self):
        self._buffers = {}
        self._csv_writers = {}
        self._files = {}

        for name, csv_headers in self.headers.items():
            self._buffers[name] = io.StringIO()
            self._csv_writers[name] =\
                csv.writer(self._buffers[name], lineterminator='\n')
            self._csv_writers[name].writerow(csv_headers)

    def _add_records(self, name, records, sortkey=None):
        if sortkey:
            records = sorted(records, key=lambda x: x[sortkey])
        for rec in records:
            if not isinstance(rec, dict):
                record = rec._asdict()
            else:
                record = rec
            csv_record = OrderedDict.fromkeys(self.headers[name])
            # Update csv with keys present in headers. Skip anything else.
            csv_record.update(
                {k: v for k, v in record.items() if k in self.headers[name]})
            self._csv_writers[name].writerow([v for v in csv_record.values()])

    @property
    def headers(self):
        """Map of filename to headers for every GTFS file.

        Only headers present in this map will be considered."""
        return {
            'agency': ['agency_id', 'agency_name', 'agency_url',
                       'agency_timezone', 'agency_lang', 'agency_phone',
                       'agency_fare_url', 'agency_email'],

            'stops': ['stop_id', 'stop_code', 'stop_name', 'stop_desc',
                      'stop_lat', 'stop_lon', 'zone_id', 'stop_url',
                      'location_type', 'parent_station', 'stop_timezone',
                      'wheelchair_boarding'],

            'routes': ['route_id', 'agency_id', 'route_short_name',
                       'route_long_name', 'route_desc', 'route_type',
                       'route_url', 'route_color', 'route_text_color'],

            'trips': ['route_id', 'service_id', 'trip_id', 'trip_headsign',
                      'shape_id'],

            'calendar': ['service_id', 'monday', 'tuesday', 'wednesday',
                         'thursday',
                         'friday', 'saturday', 'sunday', 'start_date',
                         'end_date'],

            'stop_times': ['trip_id', 'arrival_time', 'departure_time',
                           'stop_id',
                           'stop_sequence'],

            'shapes': ['shape_id', 'shape_pt_lat', 'shape_pt_lon',
                       'shape_pt_sequence'],

            'frequencies': ['trip_id', 'start_time', 'end_time',
                            'headway_secs']}

    def add_agencies(self, agencies):
        self._add_records('agency', agencies)

    def add_stops(self, stops):
        self._add_records('stops', stops)

    def add_routes(self, routes):
        self._add_records('routes', routes, sortkey=0)

    def add_calendar(self, weekly_schedules):
        self._add_records('calendar', weekly_schedules)

    def add_trips(self, trips):
        self._add_records('trips', trips)

    def add_stop_times(self, stop_times):
        self._add_records('stop_times', stop_times)

    def add_shapes(self, shapes):
        self._add_records('shapes', shapes)

    def add_frequencies(self, frequencies):
        self._add_records('frequencies', frequencies)

    def add_file(self, name, path):
        self._files[name] = path

    def write_zipped(self, filepath):
        """Write the GTFS feed in the given file."""
        with zipfile.ZipFile(filepath, mode='w', compression=zipfile.ZIP_DEFLATED) as zfile:
            for name, buffer in self._buffers.items():
                encoded_values = io.BytesIO(buffer.getvalue().encode('utf-8'))
                zfile.writestr('{}.txt'.format(name),
                               encoded_values.getbuffer())
            for name, path in self._files.items():
                zfile.write(path, arcname=name)

    def write_unzipped(self, destination):
        """Write GTFS text files in the given path."""
        for name, buffer in self._buffers.items():
            with open(os.path.join(destination,
                                   '{}.txt'.format(name)),
                      'w', encoding='utf-8') as file:
                file.write(buffer.getvalue())
        for name, path in self._files.items():
            shutil.copy(path, os.path.join(destination, name))



"""Tools to generate dummy GTFS feeds."""
import datetime
from math import radians, cos, sin, asin, sqrt
from collections import namedtuple, defaultdict

from o2g.osm.models import Agency


# Represents dummy GTFS data
DummyData = namedtuple('DummyData',
                       ['calendar', 'stop_times', 'trips', 'frequencies'])


def create_dummy_data(routes, stops):
    """Create `calendar`, `stop_times`, `trips` and `shapes`.

    :return: DummyData namedtuple
    """
    # Build stops per route auxiliary map
    stops_per_route = defaultdict(lambda: [])
    stops_map = {}
    for s in stops:
        stops_per_route[s.route_id].append(s)
        stops_map[s.stop_id] = s

    calendar = _create_dummy_calendar()

    trips = \
        _create_dummy_trips(
            routes,
            stops_per_route,
            calendar)

    stop_times = _create_dummy_stoptimes(trips, stops_per_route)
    frequencies = _create_dummy_frequencies(trips)

    return DummyData(calendar, stop_times, trips, frequencies)


def patch_agencies(agencies):
    """Fill the fields that are necessary for passing transitfeed checks."""
    # First return the unknown agency entry
    yield Agency(-1, 'http://hiposfer.com', 'Unknown agency', 'Europe/Berlin')

    # Then return the rest.
    for agency_id, agency_url, agency_name, agency_timezone in agencies:
        if not agency_url:
            agency_url = 'http://hiposfer.com'
        if not agency_timezone:
            # Set everything to one time zone to get rid of transitfeeds error.
            agency_timezone = 'Europe/Berlin'
        yield Agency(agency_id, agency_url, agency_name, agency_timezone)


def _create_dummy_calendar():
    return [{'service_id': 'WE', 'monday': 0, 'tuesday': 0, 'wednesday': 0,
             'thursday': 0,
             'friday': 0, 'saturday': 1, 'sunday': 1, 'start_date': 20170101,
             'end_date': 20190101},
            {'service_id': 'WD', 'monday': 1, 'tuesday': 1, 'wednesday': 1,
             'thursday': 1,
             'friday': 1, 'saturday': 0, 'sunday': 0, 'start_date': 20170101,
             'end_date': 20190101}]


def _create_dummy_trips(routes, stops_per_route, calendar):
    trips = []

    for route in routes:
        route_id = route.route_id

        if len(stops_per_route.get(route_id, [])) < 2:
            continue

        for cal_idx, cal in enumerate(calendar):
            # For the sake of simplicity, we assume 3 trips per service day.
            for idx in range(3):

                trip_id = \
                    '{sequence}{cal_idx}.{route_id}'.format(
                        sequence=idx + 1,
                        cal_idx=cal_idx,
                        route_id=route_id)

                trip = {'route_id': route_id,
                        'service_id': cal['service_id'],
                        'trip_id': trip_id,
                        'trip_headsign':
                            '[Dummy]{}'.format(route.route_long_name),
                        # Use route_id, i.e. relation_id as shape_id
                        'shape_id': route_id,
                        # Used for generating stop times.
                        'sequence': idx}
                trips.append(trip)

    return trips


def _create_dummy_stoptimes(trips, stops_per_route):
    stoptimes = []
    for trip in trips:
        # One service every 20 minutes from the base station
        first_service_time = \
            datetime.datetime(2017, 1, 1, 5, 0, 0) + \
            datetime.timedelta(minutes=20) * trip['sequence']
        stoptimes.extend(
            _create_dummy_trip_stoptimes(
                trip['trip_id'],
                stops_per_route.get(trip['route_id'], []),
                first_service_time))

    return stoptimes


def _create_dummy_trip_stoptimes(trip_id, stops, first_service_time):
    """Create station stop times for each trip."""
    waiting = datetime.timedelta(seconds=30)
    arrival = first_service_time
    last_departure = first_service_time
    last_departure_hour = (arrival + waiting).hour
    last_stop = None

    departure_hour = None
    arrival_hour = None
    for stop_sequence, stop in enumerate(stops):

        # Avoid time travels
        arrival = last_departure + get_time_from_last_stop(last_stop, stop)
        departure = arrival + waiting

        # Cover the case when the arrival time falls into the next day
        if arrival.hour < last_departure_hour:
            diff = last_departure_hour
            arrival_hour = arrival.hour + diff
            departure_hour = departure.hour + diff
            last_departure_hour = departure.hour + diff
        else:
            arrival_hour = arrival.hour
            departure_hour = departure.hour
            last_departure_hour = departure.hour

        # Cover the case when adding waiting time to the arrival time
        # falls into the next day
        if departure.hour < arrival.hour:
            diff = last_departure_hour
            departure_hour = departure.hour + diff
            last_departure_hour = departure.hour + diff

        yield {'trip_id': trip_id,
               'arrival_time': '{:02}:{}'.format(
                   arrival_hour,
                   arrival.strftime('%M:%S')),
               'departure_time': '{:02}:{}'.format(
                   departure_hour,
                   departure.strftime('%M:%S')),
               'stop_id': stop.stop_id,
               'stop_sequence': stop_sequence}

        last_stop = stop
        last_departure = departure


def get_time_from_last_stop(src_stop, dst_stop):
    if not src_stop:
        return datetime.timedelta()

    # Average public transport speed
    average_speed_kmh = 20

    distance_km = \
        haversine(src_stop.stop_lon,
                  src_stop.stop_lat,
                  dst_stop.stop_lon,
                  dst_stop.stop_lat)
    return \
        datetime.timedelta(
            hours=distance_km/average_speed_kmh)


# https://stackoverflow.com/a/4913653
def haversine(lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two
    points on the earth (specified in decimal degrees)
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2)**2 + cos(lat1) * cos(lat2) * sin(dlon / 2)**2
    c = 2 * asin(sqrt(a))
    r = 6371  # Radius of earth in kilometers. Use 3956 for miles
    return c * r


def _create_dummy_frequencies(trips):
    """Create station stop frequencies."""
    for trip in trips:
        yield {'trip_id': trip['trip_id'],
               'start_time': '04:30:00',
               'end_time': '08:30:00',
               'headway_secs': '1200'}
        yield {'trip_id': trip['trip_id'],
               'start_time': '08:30:00',
               'end_time': '20:30:00',
               'headway_secs': '1800'}
        yield {'trip_id': trip['trip_id'],
               'start_time': '20:30:00',
               'end_time': '25:30:00',
               'headway_secs': '2800'}



"""Functionality to build a list of stops."""
from o2g.osm.models import Stop


def build_stops(relations, nodes):
    visited_stops_ids = set()
    for rel in relations.values():
        for stop in extract_stops(rel, nodes, visited_stops_ids):
            if stop:
                yield stop


def extract_stops(relation, nodes, visited_stop_ids):
    """Extract stops in a relation."""
    for member_type, member_id, member_role in relation.member_info:

        if member_id not in visited_stop_ids and \
            member_id in nodes and\
                _is_stop(member_id, member_role, nodes):

            visited_stop_ids.add(member_id)
            yield Stop(
                member_id,
                nodes[member_id].tags.get('name') or
                "Unnamed {} stop.".format(relation.tags.get('route')),
                nodes[member_id].lon if member_id in nodes else '',
                nodes[member_id].lat if member_id in nodes else '',
                relation.id)


def _is_stop(member_id, member_role, nodes):
    """Check wether the given member designates a public transport stop."""
    return (member_role in ('stop', 'platform')) or \
        (nodes[member_id].tags.get('public_transport') == 'stop_position')



"""o2g web interface"""
import os
import pathlib
import tempfile
import urllib
from urllib.parse import urlparse
from urllib.request import urlopen

from bottle import run, template, request, static_file, abort, default_app

from o2g.cli import main


UPLOAD_DIR = '/tmp'
app = default_app()


@app.get('/')
def index():
    return template('index.html', messages=[])


@app.post('/')
def do_index():
    filepath = \
        save_file(request.files['file']) if 'file' in request.files else None

    url = request.forms.get('url')

    if not url and not filepath:
        return template('index.html', messages=['Provide a URL or upload a file.'])

    if not filepath and not is_valid_url(url):
        return template('index.html', messages=['Not a valid URL.'])

    if not filepath:
        dl_filename, dl_filepath = dl_osm_from_url(url)

    filename = os.path.split(filepath)[-1] if filepath else dl_filename
    return send_feed(
        filepath or dl_filepath,
        filename + '.gtfs.zip',
        dummy=bool(request.forms.get('dummy')))


def is_valid_url(url):
    return urlparse(url).scheme in ('http', 'https')


def save_file(file):
    if file:
        filename = file.filename
        filepath = os.path.join(UPLOAD_DIR, filename)
        file.save(filepath, overwrite=True)
        return filepath


def send_feed(filepath, filename=None, dummy=False):
    zipfile = create_zipfeed(filepath, dummy)
    if not filename:
        filename = os.path.split(zipfile)[-1]

    return static_file(
        os.path.split(zipfile)[-1],
        root=os.path.split(zipfile)[0],
        download=filename,
        mimetype='application/zip')


def create_zipfeed(filename, dummy=False):
    zipfile = '{}.zip'.format(filename)
    print('Writing GTFS feed to %s' % zipfile)
    main(filename, '.', zipfile, dummy)
    return zipfile


@app.get('/o2g', methods=['GET'])
def o2g():
    area = request.params.get('area')
    bbox = request.params.get('bbox')
    url = request.params.get('url')

    if area or bbox:
        filename, filepath = dl_osm_from_overpass(area, bbox)
        if area:
            filename = area + '_overpass'
    elif url and is_valid_url(url):
        filename, filepath = dl_osm_from_url(url)
    else:
        abort(400)

    return send_feed(
        filepath,
        filename + '.gtfs.zip',
        dummy=bool(request.params.get('dummy')))


def dl_osm_from_overpass(area, bbox):
    if not area and not bbox:
        raise Exception('At lease area or bbox must be given.')

    overpass_query = build_overpass_query(area, bbox)
    overpass_api_url = "http://overpass-api.de/api/interpreter"

    filepath = tempfile.mktemp(suffix='_overpass.osm')
    resp = urlopen(overpass_api_url, overpass_query.encode('utf-8'))

    if resp.status == 200:
        with open(filepath, 'w') as osm:
            osm.write(resp.read().decode('utf-8'))
        return os.path.split(filepath)[-1], filepath
    else:
        raise urllib.request.HTTPError(
            'Error calling Overpass API. Status: {}, reason: {}'.format(
                resp.status, resp.reason))

    raise Exception("Can't download data form overpass api.")


def build_overpass_query(area, bbox):
    template = """
    {bbox}
    {area}
    (
      rel
        [!"deleted"]
        ["type"="route"]
        ["route"~"tram|subway|bus|ex-bus|light_rail|rail|railway"]
        {area_limit};
    );
    out body;

    way(r){area_limit};
    out body;

    node(w){area_limit};
    out body;
    """
    bbox_fmt = ''
    area_fmt = ''
    area_limit_fmt = ''

    if bbox:
        south, west, north, east = bbox.split(',')
        bbox_fmt = '[bbox:{},{},{},{}];'.format(south, west, north, east)

    if area:
        area_fmt = 'area["name"="{}"]->.searchArea;'.format(area)
        area_limit_fmt = '(area.searchArea)'

    return template.format(bbox=bbox_fmt, area=area_fmt, area_limit=area_limit_fmt)


def dl_osm_from_url(url):
    filename = tempfile.mktemp(suffix=pathlib.Path(url).name)
    filepath, headers =\
        urllib.request.urlretrieve(url, filename=filename)
    print(filepath, headers)
    return pathlib.Path(url).name, filepath


if __name__ == '__main__':
    run(host='0.0.0.0', port=int(os.getenv('PORT', 3000)), debug=True)



import logging

from o2g.osm.handlers import RelationHandler, NodeHandler, WayHandler
from o2g.osm.builders import build_routes, build_stops, build_agencies,\
    build_shapes


class TransitDataExporter(object):
    def __init__(self, filename):
        self.filename = filename
        self.rh = None
        self.nh = None
        self.wh = None

    @property
    def agencies(self):
        return build_agencies(self.rh.relations, self.nh.nodes, self.wh.ways)

    @property
    def routes(self):
        return build_routes(self.rh.relations)

    @property
    def stops(self):
        return build_stops(self.rh.relations, self.nh.nodes)

    @property
    def shapes(self):
        return build_shapes(self.rh.relations, self.nh.nodes, self.wh.ways)

    def process(self):
        """Process the files and collect necessary data."""

        # Extract relations
        self.rh = RelationHandler()
        self.rh.apply_file(self.filename)

        logging.debug('Found %d public transport relations.', len(self.rh.relations))

        # Collect ids of interest
        node_ids, stop_node_ids, way_ids, reverse_map = self.__collect_ids()

        # Extract nodes
        self.nh = NodeHandler(node_ids)
        self.nh.apply_file(self.filename, locations=True)

        count = 0
        for idx, missing_node_id in enumerate(self.nh.missing_node_ids):
            count += 1
            logging.warning(
                '[no data] stop node: rel %s ref %s.',
                reverse_map[missing_node_id], missing_node_id)

        if count:
            logging.warning(
                '%d nodes that appear in relations are missing.',
                count)
        else:
            logging.debug('Lucky you! All relation member nodes were found.')

        # Extract ways
        self.wh = WayHandler(way_ids)
        self.wh.apply_file(self.filename, locations=True)

    def __collect_ids(self):
        node_ids = set()
        stop_node_ids = set()
        way_ids = set()
        reverse_map = {}

        for rel in self.rh.relations.values():
            for mtype, ref, role in rel.member_info:
                reverse_map[ref] = rel.id

                if mtype in ['n', 'node']:
                    node_ids.add(ref)
                    if role in ['stop', 'platform']:
                        stop_node_ids.add(ref)

                elif mtype in ['w', 'way']:
                    way_ids.add(ref)

                elif mtype in ['r', 'relation']:
                    logging.warning(
                        '[Rel: %s]: super-relations are not supported yet. ref: %s',
                        rel.id, ref)
                else:
                    logging.warning(
                        '[Rel: %s]: unknown member type %s, ref: %s',
                        rel.id, mtype, ref)

        return node_ids, stop_node_ids, way_ids, reverse_map



import os
import pathlib
import tempfile
import subprocess

import pytest

from o2g.osm.exporter import TransitDataExporter
from o2g.gtfs.gtfs_writer import GTFSWriter
from o2g.gtfs import gtfs_dummy


@pytest.fixture
def transit_data(request):
    root_dir = pathlib.Path(__file__).parents[2]
    filepath = root_dir.joinpath('resources', 'osm', 'freiburg.osm.bz2')
    tde = TransitDataExporter(str(filepath))
    tde.process()
    return tde


@pytest.fixture
def dummy_transit_data(transit_data):
    return \
        gtfs_dummy.create_dummy_data(transit_data.routes,
                                     transit_data.stops)


@pytest.fixture
def gtfs_writer(transit_data):
    w = GTFSWriter()

    w.add_agencies(transit_data.agencies)
    w.add_stops(transit_data.stops)
    w.add_routes(transit_data.routes)

    return w


@pytest.fixture
def dummy_gtfs_writer(transit_data, dummy_transit_data):
    w = GTFSWriter()

    # Patch agencies to pass transitfeed check
    patched_agencies = gtfs_dummy.patch_agencies(transit_data.agencies)

    w.add_agencies(patched_agencies)
    w.add_stops(transit_data.stops)
    w.add_routes(transit_data.routes)
    w.add_shapes(transit_data.shapes)

    w.add_trips(dummy_transit_data.trips)
    w.add_stop_times(dummy_transit_data.stop_times)
    w.add_calendar(dummy_transit_data.calendar)

    return w


@pytest.fixture
def dummy_zipfeed(dummy_gtfs_writer):
    filename = '{}.zip'.format(tempfile.mktemp())
    print('Writing GTFS feed to %s' % filename)
    dummy_gtfs_writer.write_zipped(filename)

    return filename


def test_write_zipped(gtfs_writer):
    filename = tempfile.mktemp()
    print('Writing GTFS feed to %s' % filename)
    gtfs_writer.write_zipped(filename)

    assert os.path.exists(filename)


def test_duplicate_trips(dummy_transit_data):
    dups = []
    for trip in dummy_transit_data.trips:
        assert trip['trip_id'] not in dups
        dups.append(trip['trip_id'])


def test_validation(dummy_zipfeed):
    """Run transitfeed over the generated feed."""
    # transitfeed is a python2 application. So we need to run it outside
    # python3 process. Moreover, it is not available in pip repository.
    # Therefore we have to clone it from git and eventually run it in a
    # process of its own. Finally we parse the standard output and look
    # for errors. We ignore the warnings for now.
    if not os.path.exists('transitfeed'):
        subprocess.check_call(
            ['git', 'clone', '-b', '1.2.16', '--single-branch',
             'https://github.com/google/transitfeed'])

    assert os.path.exists('transitfeed/feedvalidator.py')

    p = subprocess.Popen(['python2.7', 'transitfeed/feedvalidator.py', '-n',
                         dummy_zipfeed],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    out, err = p.communicate()

    print("Google Transitfeed's output:\n{}".format(out.decode('utf8')))

    assert 'error' not in out.decode('utf8')
    assert 'errors' not in out.decode('utf8')


def test_shape_id_in_trips(transit_data, dummy_transit_data):
    shape_ids = [shape.shape_id for shape in transit_data.shapes]
    for trip in dummy_transit_data.trips:
        assert trip['shape_id'] in shape_ids



import logging
import osmium as o

from o2g.osm.models import Way, Point


class WayHandler(o.SimpleHandler):
    def __init__(self, way_ids):
        super(WayHandler, self).__init__()
        self.way_ids = way_ids
        self.ways = {}

    def way(self, w):
        """Process each way."""
        if w.id not in self.way_ids:
            return

        way_points = []
        for n in w.nodes:
            try:
                way_points.append(Point(n.location.lon, n.location.lat))
            except o.InvalidLocationError:
                logging.debug('InvalidLocationError at way %s node %s', w.id, n.ref)

        self.ways[w.id] = Way(w.id, way_points)



"""Functionality to build a list of agencies."""
import hashlib

# from timezonefinder import TimezoneFinder

from o2g.osm.models import Agency


# TZ_FINDER = TimezoneFinder()


def build_agencies(relations, nodes, ways):
    extracted = []
    for rel in relations.values():
        agency = build_agency(rel, nodes)
        if agency and agency.agency_id not in extracted:
            extracted.append(agency.agency_id)
            yield agency#._replace(agency_timezone=_guess_timezone(rel, nodes, ways))


def build_agency(relation, nodes):
    """Extract agency information."""
    # TODO: find out the operator for routes without operator tag.
    # See: http://wiki.openstreetmap.org/wiki/Key:operator
    # Quote from the above link:
    #
    #    If the vast majority of a certain object in an area is operated by a certain
    #    organization and only very few by others then it may be sufficient to only tag the
    #    exceptions. For example, when nearly all roads in an area are managed by a local
    #    authority then it would be sufficient to only tag those that are not with an operator
    #    tag.
    op = relation.tags.get('operator')
    agency_url = relation.tags.get('url') or relation.tags.get('contact_website')

    if not op:
        return

    agency_id = int(hashlib.sha256(op.encode('utf8')).hexdigest(), 16) % 10**8

    return Agency(agency_id, agency_url, op, '')


# def _guess_timezone(relation, nodes, ways):
#     """Guess timezone of the relation by looking at it's nodes."""
#     lon, lat = _get_first_coordinate(relation, nodes, ways)
#     timezone = TZ_FINDER.timezone_at(lng=lon, lat=lat)
#     if not timezone:
#         logging.debug('No timezone found for (%s, %s)', lon, lat)
#     return timezone


# def _get_first_coordinate(relation, nodes, ways):
#     for member_id, member_role in relation.member_info:
#         if member_id in nodes:
#             return nodes[member_id].lon, nodes[member_id].lat
#         elif member_id in ways:
#             return _get_first_way_coordinate(member_id, ways)
#     logging.debug('No node found for relation %s', relation.id)
#     return 0, 0


# def _get_first_way_coordinate(way_id, ways):
#     # Pick the first node
#     return ways[way_id].points[0]



import logging

import osmium as o

from o2g.osm.models import Node


class NodeHandler(o.SimpleHandler):
    def __init__(self, node_ids):
        super(NodeHandler, self).__init__()
        self.node_ids = node_ids
        self.nodes = {}

    @property
    def missing_node_ids(self):
        """Get a list of nodes not found in OSM data."""
        present_node_ids = self.nodes.keys()
        for nid in self.node_ids:
            if nid not in present_node_ids:
                yield nid

    def node(self, n):
        """Process each node."""
        if n.id not in self.node_ids:
            return

        try:
            self.nodes[n.id] =\
              Node(n.id,
                   n.location.lon,
                   n.location.lat,
                   # Instead of {t.k:t.v for t in n.tags} we only pick the tags that we need,
                   # because this way it is way faster. See Profiling in README.
                   {'name': n.tags.get('name'), 'public_transport': n.tags.get('public_transport')})
        except o.InvalidLocationError:
            logging.debug('InvalidLocationError at node %s', n.id)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['o2g',
 'o2g.gtfs',
 'o2g.osm',
 'o2g.osm.builders',
 'o2g.osm.handlers',
 'o2g.tests']

package_data = \
{'': ['*']}

install_requires = \
['osmium']

extras_require = \
{'test': ['pytest'], 'web': ['bottle']}

entry_points = \
{'console_scripts': ['o2g = o2g.cli:cli']}

setup(name='o2g',
      version='0.5.1',
      description='A simple tool to extract GTFS feed from OpenStreetMap.',
      author='Mehdi Sadeghi',
      author_email='mehdi@mehdix.rog',
      url='https://github.com/hiposfer/o2g',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.5',
     )



import osmium as o

from o2g.osm.models import Relation


class RelationHandler(o.SimpleHandler):
    def __init__(self):
        super(RelationHandler, self).__init__()
        self.relations = {}
        self.versions = {}

    @property
    def transit_route_types(self):
        """A list of default OSM route types.

        Mainly train and bus types.
        """
        return ['bus', 'ex-bus',        # Bus
                'tram', 'light_rail',   # Tram
                'subway',               # Subway
                'rail', 'railway']      # Rail

    def relation(self, rel):
        """Process each relation."""
        if any([rel.deleted,
                not rel.visible,
                not self.is_new_version(rel),
                rel.tags.get('type') != 'route']):
            return

        route_tag = rel.tags.get('route')
        if route_tag not in self.transit_route_types:
            return

        self.relations[rel.id] = \
            Relation(rel.id, {
                'route': route_tag,
                'operator': rel.tags.get('operator'),
                'color': rel.tags.get('color'),
                'ref': rel.tags.get('ref'),
                'from': rel.tags.get('from'),
                'to': rel.tags.get('to'),
                'name': rel.tags.get('name'),
                'alt_name': rel.tags.get('alt_name'),
                'url': rel.tags.get('url'),
                'contact_website': rel.tags.get('contact:website')},
                [(member.type, member.ref, member.role) for member in rel.members])

        self.versions[rel.id] = rel.version

    def is_new_version(self, relation):
        return relation.id not in self.versions or \
            relation.version > self.versions[relation.id]



"""Functionality to build a list of routes."""
import hashlib

from o2g.osm.models import Route
from o2g.gtfs.gtfs_misc import map_osm_route_type_to_gtfs


def build_routes(relations):
    for rel in relations.values():
        route = build_route(rel)
        if route:
            yield route


def build_route(relation):
    """Extract information of one route."""
    short_name = create_route_short_name(relation)
    color = relation.tags.get('color')
    return\
        Route(relation.id,
              short_name,
              create_route_long_name(relation, short_name),
              map_osm_route_type_to_gtfs(relation.tags.get('route')),
              'https://www.openstreetmap.org/relation/{}'.format(relation.id),
              color.strip('#') if color else '',
              get_agency_id(relation))


def create_route_short_name(relation):
    """Create a meaningful route short name."""
    return relation.tags.get('ref') or ''


def create_route_long_name(relation, short_name):
    """Create a meaningful route name."""
    if relation.tags.get('from') and relation.tags.get('to'):
        return "{0}-to-{1}".format(relation.tags.get('from'),
                                   relation.tags.get('to'))
    name = relation.tags.get('name') or\
        relation.tags.get('alt_name') or\
        "OSM Route No. {}".format(relation.id)

    # Drop route_short_name from this one if it contains it
    if short_name and name.startswith(short_name):
        # Drop it
        return name[len(short_name):]
    return name


def get_agency_id(relation):
    """Construct an id for agency using its tags."""
    op = relation.tags.get('operator')
    if op:
        return int(hashlib.sha256(op.encode('utf-8')).hexdigest(), 16) % 10**8
    return -1



"""Functionality to build a list of shapes."""
import logging

from o2g.osm.models import Shape


def build_shapes(relations, nodes, ways):
    for rel in relations.values():
        for record in build_shape(rel, nodes, ways):
            if record:
                yield record


def build_shape(relation, nodes, ways):
    """Extract shape of one route."""
    sequence_index = 0

    for member_type, member_id, member_role in relation.member_info:

        if member_id in nodes:
            yield Shape(
                relation.id,
                nodes[member_id].lat,
                nodes[member_id].lon,
                sequence_index)

            sequence_index += 1

        # Do we need to consider ways too? It dramatically increases the number of shapes.
        elif member_id in ways:
            continue
        #     for point in ways[member_id].points:
        #         shape = Shape(
        #             relation.id,
        #             point.lat,
        #             point.lon,
        #             sequence_index)
        #         sequence_index += 1

        else:
            logging.warning('[no data] rel %s type %s ref %s.',
                            relation.id, member_type, member_id)
